"""@package docstring
This module contains the Schema class.
TODO: More info
"""

from typing import Dict, List

from b_asic.signal_flow_graph import SFG
from b_asic.graph_component import GraphID
from b_asic.operation import Operation


class Schema:
    """A class that represents an SFG with scheduled Operations."""

    _sfg: SFG
    _start_times: Dict[GraphID, int]
    _laps: Dict[GraphID, List[int]]
    _schedule_time: int
    _cyclic: bool
    _resolution: int

    def __init__(self, sfg: SFG, schedule_time: int = None, cyclic: bool = False, resolution: int = 1, scheduling_alg: str = "ASAP"):

        self._sfg = sfg
        self._start_times = dict()
        self._laps = dict()
        self._cyclic = cyclic
        self._resolution = resolution
        self._memory_elements = dict()

        if scheduling_alg == "ASAP":
            self._schedule_asap()
        else:
            raise NotImplementedError(f"No algorithm with name: {scheduling_alg} defined.")

        max_end_time = 0
        for op_id, op_start_time in self._start_times.items():
            op = self._sfg.find_by_id(op_id)
            for outport in op.outputs:
                max_end_time = max(max_end_time, op_start_time + outport.latency_offset)

        if not self._cyclic:
            if schedule_time is None:
                self._schedule_time = max_end_time
            elif schedule_time < max_end_time:
                raise ValueError("Too short schedule time for non-cyclic Schedule entered.")
            else:
                self._schedule_time = schedule_time

        self.get_memory_elements()

    def start_time_of_operation(self, op_id: GraphID):
        """Get the start time of the operation with the specified by the op_id."""
        assert op_id in self._start_times, "No operation with the specified op_id in this schema."
        return self._start_times[op_id]

    def forward_slack(self, op_id):
        raise NotImplementedError

    def backward_slack(self, op_id):
        raise NotImplementedError

    def print_slacks(self):
        raise NotImplementedError

    def _schedule_asap(self):
        pl = self._sfg.get_precedence_list()

        if len(pl) < 2:
            print("Empty signal flow graph cannot be scheduled.")
            return

        non_schedulable_ops = set((outp.operation.graph_id for outp in pl[0]))

        for outport in pl[1]:
            op = outport.operation
            if op not in self._start_times:
                # Set start time of all operations in the first iter to 0
                self._start_times[op.graph_id] = 0

        for outports in pl[2:]:
            for outport in outports:
                op = outport.operation
                if op.graph_id not in self._start_times:
                    # Schedule the operation if it doesn't have a start time yet.
                    op_start_time = 0
                    for inport in op.inputs:
                        print(inport.operation.graph_id)
                        assert len(inport.signals) == 1, "Error in scheduling, dangling input port detected."
                        assert inport.signals[0].source is not None, "Error in scheduling, signal with no source detected."
                        source_port = inport.signals[0].source

                        source_end_time = None
                        if source_port.operation.graph_id in non_schedulable_ops:
                            source_end_time = 0
                        else:
                            source_op_time = self._start_times[source_port.operation.graph_id]

                            assert source_port.latency_offset is not None, f"Output port: {source_port.index} of operation: \
                                    {source_port.operation.graph_id} has no latency-offset."
                            assert inport.latency_offset is not None, f"Input port: {inport.index} of operation: \
                                    {inport.operation.graph_id} has no latency-offset."

                            source_end_time = source_op_time + source_port.latency_offset

                        op_start_time_from_in = source_end_time - inport.latency_offset
                        op_start_time = max(op_start_time, op_start_time_from_in)

                    self._start_times[op.graph_id] = op_start_time

    def get_memory_elements(self):
        pl = self._sfg.get_precedence_list()

        for port_list in pl:
            for port in port_list:
                time_list = []
                for key in self._start_times:
                    if port.operation.graph_id == key:
                        for i in range(len(port.operation.outputs)-1):
                            time_list.append(self.start_times.get(port.operation.graph_id)+op.outputs[i].latency_offset)
                            for j in range(len(port.operation.outputs[i].signals)-1):
                                print(self._start_times.get(port.operation.outputs[i].signals[j].destination.operation.op_id))
                                time_list.append(self._start_times.get(port.operation.outputs[i].signals[j].destination.operation.op_id))
                                self._memory_elements[port.operation.outputs[i].signals[j].type_name] = time_list
                                time_list.pop() # remove end time if an output port has several signals as it's source

    def print_memory_elements(self):
        for key in self._memory_elements:
            for value in self._memory_elements[key]: 
                print(key, value)


